PROGRAM vigenere; {$M 20000, 10000}

{ Vigenere-Verfahren mit Friedman-Test, C. Baensch 1998 mit TPW 1.5 fuer LOGIN

  Zur DATENABSTRAKTION: Fuer Ein- und Ausgabe ist es einfacher, einen Text als
  Feld t aus z Zeilen zu behandeln, zum Chiffrieren und Dechiffrieren ist er
  als Feld c aus n Zeichen leichter handhabbar. Der Text existiert deswegen 
  waehrend der Bearbeitung in diesen zwei Formen, so dass bis zu vier Parameter
  uebergeben werden: t und z, c und n.

  Die BENUTZERSCHNITTSTELLE ist einfach gehalten, da sie hier zweitrangig ist.
  Man kann sie komfortabler machen und so den Code um 100 Zeilen aufblaehen.
  Er ist so schon lang genug: ca. 400 LoC.

  Die INLINE-DOKUMENTATION soll das Verstaendnis des Programms erleichtern.
  Sie ist die einzige Dokumentationsart, die nicht verloren geht und die man
  wirklich liest, wenn man sich fuer den Quellcode interessiert.              }

USES  WinCrt, WinDos, Strings;

CONST zl = 75;                             {max. Zeilenlaenge einer Textzeile}
      zz = 25;                                  {max. Zeilenzahl eines Textes}
      dl = 12;                   {max. Dateinamenlaenge gemaess 8.3-Konvention}
      xe = 46;                               {Eingabeposition fuer den Dialog}

TYPE  tDName = string[dl];                       {Dateiname in 8.3-Konvention} 
      tText  = array [1..zz] of string;                  {Text als Zeilenfeld}
      tCFeld = array [1..zl*zz] of char;                {Text als Zeichenfeld}
      tCSet  = set of char;             {Zeichenvorrat bei Bildschirmeingaben}

{--------------------------- Hilfsroutinen ----------------------------------}

PROCEDURE init (xol,yol,br,h: integer; titel: PChar);
  {initialisiert das Crt-Fenster}
BEGIN
  WindowOrg.x := xol;
  WindowOrg.y := yol;
  ScreenSize.x:=  br;
  ScreenSize.y:=   h;
  StrCopy (WindowTitle, titel);
  InitWinCrt
END; {init}

FUNCTION lies_int (ugr, ogr: integer): integer;
  {Einlesen einer max. 8-stelligen Ganzzahl im Bereich zwischen ugr und ogr}
VAR n, x,y, code: integer;
    s           : string[8];
BEGIN
  x := wherex; y := wherey;
  REPEAT
    gotoxy(x,y); readln (s);                             {Einlesen als String}
    val(s,n, code);                                    {Umwandeln in Ganzzahl}
    IF   (n<ugr) OR (n>ogr) OR (code <> 0)
    THEN BEGIN
           gotoxy (x,y); ClrEol
         END;
  UNTIL code = 0;
  lies_int := n
END; {fct. liesint}

FUNCTION antwort (frage: string; xe: byte; vorrat: tCSet): char;
  {gibt frage aus, liest an der Pos. y ein Zeichen ohne Enter ein, Z.Vorschub}
VAR a: char;
    x: byte;
BEGIN
  write (frage, ' ':xe-length(frage));
  x := wherex;                       {nicht ye, fallse die Frage zu lang ist!}
  ClrEol; gotoxy (x, wherey);
  REPEAT
    a := readkey
  UNTIL a in vorrat;
  writeln (a);
  antwort := a
END; {fct. antwort}

PROCEDURE c_to_z (VAR t: tText; VAR z: integer; c: tCFeld; n: integer);
  {Wandelt ein Zeichenfeld um in ein Zeilenfeld mit maximierten Zeilenlaengen}
VAR i : integer;
BEGIN
  z := n div zl +1;                                 {Zeilenzahl neu berechnen}
  FOR i := 1 TO z DO t[i] := '';                        {t neu initialisieren}
  FOR i := 1 TO n
  DO  t[(i-1)div zl +1] := t[(i-1)div zl +1] + c[i]           {t neu aufbauen}
END; {c_to_z}

PROCEDURE aufbereiten (VAR t: tText; VAR z: integer;
                       VAR c: tCFeld;VAR n: integer);
  {Wandelt den Text von einem Zeilen- in ein Zeichenfeld nur aus den Zeichen
   'a'..'z' um und reorganisiert zurueck in maximierte Zeilen, nutzt c_to_z}
VAR i, j: integer;
BEGIN
  n := 0;                                    {t in ein Zeichenfeld c aus a..z}
  FOR i := 1 TO z
  DO  FOR j := 1 TO length (t[i])
      DO BEGIN
           IF   t[i][j] IN ['A'..'Z']                               {downcase}
           THEN t[i][j] := chr(ord(t[i][j])+32);
           IF   t[i][j] IN ['a'..'z']              {alle anderen Zeichen raus}
           THEN BEGIN
                  inc (n);
                  c[n] := t[i][j]
                END
         END;
  c_to_z (t,z, c,n)
END; {aufbereiten}

{-------------------------- In-Out-Routinen ---------------------------------}

PROCEDURE speichern (VAR d: tDname; t: tText; z: integer);
  {Speichern eines Textes in einer Textdatei}
VAR a   : char;
    f   : text;
    i   : 1..zz;
    dalt: tDName;
BEGIN
  dalt := d;
  REPEAT
    a := 'j';
    write ('Speichern unter (Abbruch mit "x"):',' ':xe-34);
    readln (d);
    IF   d <> 'x'
    THEN BEGIN
           assign (f, d);
           {$I-} reset (f); {$I+}                 {Programmabbruch verhindern}
           IF   ioresult = 0
           THEN a := antwort ('Datei existiert schon. Ueberschreiben (j/n)?',
                              xe,['j','n'])
         END
  UNTIL a = 'j';
  IF   d <> 'x'
  THEN BEGIN
         writeln('Der Text wird als "', d, '" gespeichert.');  
         assign (f, d);
         rewrite(f);
         FOR i := 1 TO z DO writeln (f,t[i]);
         close (f)
       END
  ELSE d := dalt
END; {speichern}

PROCEDURE ausgeben (t: tText; z: integer; d: tDName);
  {Bildschirm- und ggf. DOS-Druckerausgabe des Textes t}
VAR i  : integer;
    a  : char;
    dev: text;
BEGIN
  writeln;
  FOR i := 1 TO z DO writeln (i:2, ' ',t[i]);
  writeln;
  a := antwort ('Text drucken (j/n)?', xe, ['j','n']);
  IF   a = 'j'
  THEN BEGIN
         assign  (dev, 'prn');
         rewrite (dev);
         writeln (dev);
         writeln (dev, d);
         FOR i := 1 TO z DO writeln (dev, i:2,' ',t[i]);
         writeln (dev);
         close   (dev)
       END
END; {ausgeben}

PROCEDURE schreiben (VAR t: tText; VAR z: integer);
  {Simple Texteingabe ohne weitere Editierfunktionen}
VAR i: integer;
BEGIN
  writeln ('Klartext schreiben (Am Zeilenende 1x, am Textende 2x <Enter>)');
  writeln ('(Fuer Umlaute und "sz" Ersatzzeichen ae,oe,ue,ss verwenden.):');
  i := 0;
  REPEAT
    inc   (i);
    write (i:2,' ');
    readln(t[i])
  UNTIL (t[i] = '');
  z := pred (i)                       {Leerzeile am Ende wieder wegschmeissen}
END; {schreiben}

PROCEDURE laden (VAR d: tDname; VAR t: tText; VAR z: integer);
  {laedt eine Textdatei d in ein Zeilenfeld t mit z Zeilen}
VAR f: text;
    k: word; 
BEGIN
  REPEAT
    write  ('Dateiname (Abbruch mit "x"):', ' ':xe-28);
    readln (d);
    assign (f, d);
    {$I-} reset(f); {$I+}              {moeglichen Programmabbruch verhindern}
    k := ioresult;
    IF   (d <> 'x') AND (k <> 0)
    THEN writeln ('Die Datei ', d, ' konnte nicht geoeffnet werden.');
  UNTIL (d = 'x') OR (k = 0);
  IF   d <> 'x'
  THEN BEGIN
         z := 0;
         WHILE NOT eof(f)
         DO BEGIN
              inc (z);
              readln (f, t[z]);
            END; {while}
         close (f)
       END
END; {laden}

PROCEDURE hilfe;
  {Ausgabe eines Hilfetextes, koennte auch leicht in eine Extra-Hilfe-Datei}
VAR   i  : byte;
      h  : tText;
      a  : char;
BEGIN
  h[ 1]:='Das Vigenere-Verfahren ist ein polyalphabetisches Chiffrierverfahren.';
  h[ 2]:='Ein Text wird mit Hilfe eines Schluesselwortes zyklisch zeichenweise';
  h[ 3]:='unterschiedlich Caesar-verschluesselt. Je laenger das Schluesselwort';
  h[ 4]:='und je kuerzer der Klartext, desto sicherer ist das Verfahren.';
  h[ 5]:='Sind Text und Schluesselwort gleich lang und das Schluesselwort ein';
  h[ 6]:='Zufallstext, dann ist das Verfahren absolut sicher.';
  h[ 7]:='Um das Schluesselwort zu rekonstruieren, wird eine statistische Analyse';
  h[ 8]:='periodischer Buchstabenhaeufigkeiten durchgefuehrt (Friedman-Test).';
  h[ 9]:='So kann eine Vermutung fuer die Schluesselwortlaenge aufgestellt werden.';
  h[10]:='Als Buchstabe mit der groessten Haeufigkeit wird das verschluesselte';
  h[11]:='"e" angenommen. Nach einem Dechiffrierungsvorschlag kann jede Position';
  h[12]:='im Schluesselwort einzeln sukzessive korrigiert werden.';
  h[13]:='Fuer dieses Programm gilt:';
  h[14]:='- Ein Text darf max. aus 25 Zeilen zu je max. 75 Zeichen bestehen.';
  h[15]:='- Ein Schluesselwort darf hoechstens 25 Zeichen lang sein.';
  h[16]:='- Zahl- und Stringeingaben erfolgen mit, Zeicheneingaben ohne <Enter>.';
  writeln; 
  FOR i := 1 TO 16 DO writeln (' ', h[i]);
  writeln;
  a := antwort (' Weiter mit <Enter>...', xe, [chr(13)]);
  gotoxy (1, wherey-19);
  FOR i := 1 TO 19 DO writeln (' ':zl);
  gotoxy (1, wherey-20);
END; {hilfe}

{------------------------- Vigenere-Algorithmen -----------------------------}

PROCEDURE entschluesseln (VAR t: tText; z: integer; VAR c: tCFeld; n: integer);
  {findet Korrelationen, wertet sie aus und ermoeglicht benutzergesteuerte,
   zeichenweise Modifikationen des Schluesselwortes. nutzt c_to_z}  

VAR l: integer; wort: string;

  PROCEDURE koinzidenz (c: tCFeld; n: integer);
    {Friedman-Test: Untersuchung von Korrelationen}
  VAR i, j, gr: integer;                           
      koin: array [1..100] of real;
  BEGIN
     {Test auch groesserer Abstaende, um durch Vielfache l(wort) zu ermitteln}
    IF  n > 200 THEN gr := 100 ELSE gr := trunc (n/2);
    FOR i := 1 TO gr DO koin[i] := 0;                        {Initialisierung}
    FOR i := 1 TO gr                    {Korrelationspruefung Abstaende 1..gr}
    DO FOR j := 1 TO n              {fuer beliebige uebereinstimmende Zeichen}
       DO IF   c[j] = c[(i+j)mod n]
          THEN koin[i] := koin[i]+1;
    FOR i := 1 TO gr
    DO  koin[i] := koin[i]/n;            {absolute --> relative Haeufigkeiten}
    writeln;
    writeln ('Zeichenabstaende mit Koinzidenz groesser als 8%:');
    FOR i := 1 TO gr
    DO  IF   koin[i] > 0.08   {i.d.R. reichen 8%, mehr kleine Werte verwirren}
        THEN writeln (i:4, 100*koin[i]:6:1,' %')
  END; {koinzidenz}

  PROCEDURE dechiff (VAR c:tCFeld; n:integer; VAR l:integer; VAR wort:string);
    {ermittelt die wahrscheinlichste Verschiebung durch Vergleich der Position
     des haeufigsten Zeichens in jedem der l Teiltexte mit e und rekonstruiert
     so das Schluesselwort. Fehler geschehen, wenn ein anderer Buchstabe, z.B.
     n oder s, zufaelligerweise in einem Teiltext haeufiger als e auftritt}
   VAR i, j    : integer;
       h       : array [1..26] of array [1..26] of integer;
       max, ind: array [1..26] of integer;
   BEGIN
    write ('Vermutete Schluesselwortlaenge (ggT(koinz))?   ');
    l := lies_int (1,26);
    writeln ('Haeufigkeitsanalyse der ',l:1,' Teiltexte');
    FOR i := 1 TO l                           {1. Index: Schluesselwortstelle}
    DO  FOR j := 1 TO 26                       {2. Index: ord(Buchstabe a..z)}
        DO  h[i][j] := 0;
    FOR i := 1 TO n                             {abs. Haeufigkeiten ermitteln}
    DO  h[(i-1)mod l+1][ord(c[i])-96] := succ(h[(i-1)mod l+1][ord(c[i])-96]);
    FOR i := 1 TO 26 DO write (chr(i+96):3);                    {Ueberschrift}
    writeln;
    FOR i := 1 TO 78 DO write ('-');
    writeln;
    FOR i := 1 TO l               {abs. Haeufigkeiten fuer l Stellen ausgeben}
    DO BEGIN
         FOR j := 1 TO 26
         DO  write (h[i][j]:3);
         writeln
       END;
    FOR i := 1 TO l                                             {Maximumsuche}
    DO BEGIN
         max[i] := 0;
         FOR j := 1 TO 26
         DO  IF   h[i][j] > max[i]
             THEN BEGIN
                    max[i] := h[i][j]; ind[i] := j
                  END
       END;
    write ('Vermutete Verschiebungen der Textzeichen:    ');
    FOR i := 1 TO l DO ind[i]:=(ind[i]-5+26)mod 26; {Abstand der Mmaxima zu e}
    FOR i := 1 TO l DO write (' ', ind[i]:1);
    writeln;
    wort := '';                           {Rekonstruktion des Schluesselworts}
    FOR i := 1 TO l DO wort := wort+ chr(97+ (ind[i]+26)mod 26);
    writeln ('Vermutetes Schluesselwort:', ' ':xe-26, wort);
    writeln;
    FOR i := 1 TO n               {Entschluesselung mit wort zuerst nur auf c}
    DO  c[i] := chr((ord(c[i])-ind[(i-1) mod l+1]-19) mod 26 +97)
  END; {dechiff}

  PROCEDURE kontrolle (VAR c: tCFeld; n: integer; wort: string);
    {benutzergesteuerte Korrektur einzelner Zeichen des Schluesselwortes}
  VAR i, st, v: integer;
      ant     : char;
  BEGIN
    REPEAT
      FOR i := 1 TO n
      DO BEGIN
           write (c[i]);
           IF i mod 75 = 0 THEN writeln
         END;
      writeln; writeln;
      ant := antwort ('Korrekt so (j/n)?', xe, ['j','n']);
      IF   ant = 'n'
      THEN BEGIN
             write ('Falsche Stelle im Schluesselwort (0->Abbr.)?  ');
             st := lies_int (0, l);
             IF   st <> 0
             THEN BEGIN
                    write ('Wie gross soll dort die Aenderung sein?       ');
                    v := lies_int (-26,26);
                    wort[st] := chr ((ord(wort[st])+v-19) mod 26 + 97);
                    writeln ('Geaendertes Schluesselwort:',' ':xe-27, wort);
                    writeln;
                    FOR i := 1 TO n
                    DO  IF   (i-st) mod l = 0
                        THEN c[i] := chr((ord(c[i])-v-19) mod 26+97);
                  END {if st..}
             ELSE ant := 'j'
           END {if ant..}
    UNTIL ant = 'j'
  END; {kontrolle}

BEGIN {entschluesseln}
  koinzidenz(c,n);
  dechiff   (c,n,l,wort);
  kontrolle (c,n,wort);
  c_to_z    (t,z, c,n)
END; {entschluesseln}

PROCEDURE verschluesseln (VAR t: tText; z: integer; VAR c: tCFeld; n: integer);
  {verschluesselt den Text t mit einem Schluesselwort, nutzt c_to_z}
VAR w, wort: string;
    i, l: integer;
BEGIN
  write ('Ihr Schluesselwort (nur a-z, Laenge < 26):    ');
  readln(w);
  IF   length(w) > 25
  THEN delete (w, 26, length (w));
  l := 0;
  FOR i := 1 TO length (w)
  DO IF   w[i] IN ['a'..'z']
     THEN BEGIN
            inc (l);
            wort[l] := w[i]
          END;
  FOR i:= 1 TO n
  DO  c[i] := chr((ord(c[i])+ord(wort[(i-1)mod l +1]) -12) mod 26 +97);
  c_to_z (t,z, c,n)
END; {verschluesseln}

{-------------------------- Ablaufsteuerung ---------------------------------}

PROCEDURE anwendung;
  {Bei nur einem Verfahren lohnt sich eine Menuesteuerung nicht}
VAR a   : char;                                      {Quasi-globale Variablen}
    t   : tText;
    d   : tDName;
    z, n: integer;
    c   : tCFeld;
BEGIN
  gotoxy (1,11);
  writeln('Bitte maximieren Sie das Fenster, damit alle Textzeilen ',
          'vollstaendig sichtbar sind <Enter>');
  readln;
  ClrScr;
  d := 'x';
  REPEAT
    REPEAT
      a := antwort('Text (l)aden/(s)chreiben/(v)erwenden/(a)bbrechen/(h)ilfe?',
                   xe, ['a','h','l','s','v']);
      IF a = 'h' THEN hilfe;
      IF   (a = 'v') AND (d = 'x')
      THEN BEGIN
             writeln ('Sie muessen erst einen Text laden oder Speichern!');
             a := 'h'
           END
    UNTIL a <> 'h';
    IF   a = 'l'
    THEN BEGIN
           laden      (d, t,z);
           aufbereiten(t,z, c,n);
           IF t[1] <> '' THEN ausgeben (t,z, d)
         END;
    IF   a = 's'
    THEN BEGIN
           schreiben  (t,z);
           aufbereiten(t,z, c,n);
           ausgeben   (t,z, d);
           speichern  (d, t,z)
         END;
    IF a <> 'a'                            {Sonst soll ja abgebrochen werden.}
    THEN BEGIN
           REPEAT
             a := antwort ('(v)er-/(e)ntschluesseln/(a)bbrechen/(h)ilfe?',
                           xe, ['a','e','h','v']);
             IF a = 'h' THEN hilfe
           UNTIL a <> 'h';
           IF a = 'e' THEN entschluesseln (t,z, c,n);
           IF a = 'v'
           THEN BEGIN
                  verschluesseln(t,z, c,n);
                  ausgeben      (t,z, d)
                END;
           IF a <> 'a' THEN speichern(d, t,z);
         END; {if a...}
    a := antwort ('Weiter (j/n)?', xe, ['j','n']);
    writeln
  UNTIL a <> 'j'
END; {anwendung}

BEGIN {main}
  init (30,50,90,35,'Vigenere-Verfahren und Friedman-Test (C. Baensch 98)');
  anwendung;
  DoneWinCrt
END. {main}